"""Light classes to handle parameters for classes construction.

.. currentmodule hysop.tools

"""

from collections import namedtuple
from hysop.mpi.main_var import main_comm, main_rank, MPI
from hysop.constants import DEFAULT_TASK_ID


class MPIParams(namedtuple('MPIParams', ['comm', 'task_id',
                                         'rank', 'on_task'])):
    """
    Struct to save mpi parameters :
    - comm : parent mpi communicator (default = main_comm)
    - task_id : id of the task that owns this object
    (default = DEFAULT_TASK_ID)
    - rank of the current process in comm
    - on_task : true if the task_id of the object corresponds
    to the task_id of the current process.

    This struct is useful for operators : each operator has
    a MPIParams attribute to save its mpi settings.

    Examples
    ---------

    op = SomeOperator(..., task_id=1)
    if op.is_on_task():
       ...

    'is_on_task' will return MPIParams.on_task value for op
    and tell if the current operator belongs to the current process
    mpi task.
    """
    def __new__(cls, comm=main_comm, task_id=DEFAULT_TASK_ID,
                rank=main_rank, on_task=True):
        if comm != MPI.COMM_NULL:
            rank = comm.Get_rank()
        else:
            rank = MPI.UNDEFINED
        return super(MPIParams, cls).__new__(cls, comm, task_id,
                                             rank, on_task)


import hysop.tools.numpywrappers as npw


class Discretization(namedtuple("Discretization", ['resolution', 'ghosts'])):
    """
    A struct to handle discretization parameters:
    - a resolution (either a list of int or a numpy array of int)
    - number of points in the ghost-layer. One value per direction, list
    or array. Default = None.
    """
    def __new__(cls, resolution, ghosts=None):
        resolution = npw.asdimarray(resolution)
        if ghosts is not None:
            ghosts = npw.asintarray(ghosts)
            msg = 'Dimensions of resolution and ghosts parameters'
            msg += ' are not complient.'
            assert ghosts.size == resolution.size, msg
            assert all(ghosts >= 0)
        else:
            ghosts = npw.int_zeros(resolution.size)
        return super(Discretization, cls).__new__(cls, resolution, ghosts)

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return (self.resolution == other.resolution).all() and\
            (self.ghosts == other.ghosts).all()

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

from hysop.constants import HDF5
import os


class IOParams(namedtuple("IOParams", ['filename', 'filepath',
                                         'frequency', 'fileformat',
                                         'io_leader'])):
    """
    A struct to handle I/O files parameters

    Parameters
    -----------
    filename : string
        name of the file (absolute or relative path)
    filepath : string
        location of the file
    frequency : int
        frequency of output or input (e.g. every N times steps)
    fileformat : int
        format of the file. See notes for available format. Default=HDF5.
    io_leader : int
        rank of the mpi process dealing with the io. Default is 0.


    See examples in hysop.operator.hdf_io

    Notes
    -----
    Format parameter must be one of the following :
      - hysop.constants.HDF5
      - hysop.constants.ASCII

    """
    def __new__(cls, filename, filepath=None, frequency=1,
                fileformat=None, io_leader=0):
        from hysop.tools.io_utils import IO

        # Filename is absolute path, filepath arg is ignored.
        if os.path.isabs(filename):
            filepath = os.path.dirname(filename)

        else:
            if filepath is not None:
                filename = os.path.join(filepath, filename)
                filepath = os.path.abspath(os.path.dirname(filename))
            else:
                filepath = os.path.dirname(filename)
                if filepath == '':
                    # Get default output path
                    filepath = IO.default_path()
                    filename = os.path.join(filepath, filename)
                else:
                    filepath = os.path.abspath(filepath)
                    filename = os.path.join(filepath,
                                            os.path.basename(filename))
        if fileformat is None:
            fileformat = HDF5

        IO.check_dir(filename)
        return super(IOParams, cls).__new__(cls, filename, filepath,
                                            frequency, fileformat, io_leader)
